#include <linux/module.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/miscdevice.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/io.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/irq_vectors.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#define  BIFFIRQ_MAJOR     32
#define  BIFFIRQ_NAME      "biffirq"
#define  BIFFIRQ_DRV_NAME  "bifferboard pin12/IRQ"
#define  BIFFIRQ_PIN_NUM   12
#define  BIFFIRQ_IRQ_NUM   4

#define  PCI_ADDR_SEL      (0xcf8)
#define  PCI_DATA_REG      (0xcfc)
#define  GPIO_REG_1_CTRL   (0x80003848)

#define  MODEM_IRQ_ENABLE  (0x3f9)
#define  MODEM_IRQ_IDENT   (0x3fa)
#define  MODEM_CTRL        (0x3fc)
#define  MODEM_STATUS      (0x3fe)

#define  get_wait_flag(x, y)  do { x = 1; while (y & x) x <<= 1; } while (0)

static int _dev_open = 0;
static int _pin_val = 0;
static u32 _taken_cond_flags = 0;
static volatile u32 _pin_val_change_cond_poll = 0;
static volatile u32 _pin_val_change_cond_read = 0;

DECLARE_WAIT_QUEUE_HEAD(biffirq_wait);
static DEFINE_SPINLOCK(biffirq_lock);

static inline u32 pci_read_data(u32 reg)
{
   outl(reg, PCI_ADDR_SEL);
   return inl(PCI_DATA_REG);
}

static inline void pci_write_data(u32 reg, u32 data)
{
   outl(reg, PCI_ADDR_SEL);
   outl(data, PCI_DATA_REG);
}

static inline int set_bit_at_port_and_verify8(u32 port, u8 bit)
{
   u8 val8;
   int err = 0;
   
   val8 = inb(port);
   outb(val8 | bit, port);
   val8 = inb(port);
   if (!(val8 & bit))
      err = -ENODEV;
   
   return err;
}

static inline int clear_bit_at_port_and_verify8(u32 port, u8 bit)
{
   u8 val8;
   int err = 0;
   
   val8 = inb(port);
   outb(val8 & ~bit, port);
   val8 = inb(port);
   if ((val8 & bit))
      err = -ENODEV;
   
   return err;
}

static irqreturn_t irq_handler(int irq, void *dev_id)
{	
   int pin_val, did_change = 0;
   unsigned long irqflags;
	
	spin_lock_irqsave(&biffirq_lock, irqflags);
	
	// reset pci address select. someone else might've
   // changed it.
   outl((u32)0, PCI_ADDR_SEL);
	
	// DSR_n is active-low, so we should invert the value here,
	// as a 1 actually means it's currently low.
   pin_val = (((inb(MODEM_STATUS) >> 5) & 0x1) + 1) % 2;
   if (pin_val != _pin_val) {
      _pin_val = pin_val;
      _pin_val_change_cond_read = 0;
      _pin_val_change_cond_poll = 0;
      
      did_change = 1;
   }
   
   spin_unlock_irqrestore(&biffirq_lock, irqflags);
   
   if (did_change)
      wake_up_interruptible_all(&biffirq_wait);
	
	// return IRQ_NONE, so that the actual serial driver can
	// also see this interrupt.
	return IRQ_NONE;
}

static int configure_irq_pin(u32 pin_nr)
{
   u32 val32;
   unsigned long irqflags;
   int err = 0, locked = 0;
   
   if (BIFFIRQ_PIN_NUM != pin_nr) {
      err = -EINVAL;
      goto fault;
   }

   spin_lock_irqsave(&biffirq_lock, irqflags);
   locked = 1;
   
   // if GPIO is currently enabled on our pin, turn it off
   val32 = pci_read_data(GPIO_REG_1_CTRL);
   val32 &= ~(1 << pin_nr);
   pci_write_data(GPIO_REG_1_CTRL, val32);
   
   // read back to confirm
   val32 = pci_read_data(GPIO_REG_1_CTRL);
   if (val32 & (1 << BIFFIRQ_PIN_NUM)) {
      printk(KERN_ERR BIFFIRQ_DRV_NAME ": failed to turn off GPIO "
               "on pin %d.\n", pin_nr);
      err = -ENODEV;
      goto fault;
   }
   
   // reset pci address select
   outl((u32)0, PCI_ADDR_SEL);
   
   // set modem line status change interrupt enable and verify
   if ((err = set_bit_at_port_and_verify8(MODEM_IRQ_ENABLE, (1<<3)))) {
      printk(KERN_ERR BIFFIRQ_DRV_NAME ": failed to enable modem line "
                  "status change interrupt on pin %d.\n", pin_nr);
      goto fault;
   }

   err = request_irq(BIFFIRQ_IRQ_NUM, irq_handler,
   			IRQF_SHARED, BIFFIRQ_NAME, (void*)(&irq_handler));
   			
   if (err)
      goto fault;
   
   locked = 0;
   spin_unlock_irqrestore(&biffirq_lock, irqflags);
   
   // tell the hardware that we want the interrupt on DSR change.
   (void)set_bit_at_port_and_verify8(MODEM_STATUS, (1<<1));
   
   printk(KERN_INFO BIFFIRQ_DRV_NAME ": pin 12 configured for "
               "interrupt on value change: /dev/%s\n", BIFFIRQ_NAME);
         
fault:

   if (locked)
      spin_unlock_irqrestore(&biffirq_lock, irqflags);

   return err;
}

static int biffirq_open(struct inode *inode, struct file *file)
{
   u32 wait_flag;
   unsigned long irqflags;
   
   //try_module_get(THIS_MODULE);
   
   if (_dev_open > 31)
      return -EBUSY;
   
   spin_lock_irqsave(&biffirq_lock, irqflags);
   
   get_wait_flag(wait_flag, _taken_cond_flags);
   file->private_data = (void*)wait_flag;
   
   _taken_cond_flags |= wait_flag;
   _pin_val_change_cond_read &= ~((u32)file->private_data);
   _pin_val_change_cond_poll &= ~((u32)file->private_data);
   
   _dev_open++;
   
   spin_unlock_irqrestore(&biffirq_lock, irqflags);
   
   return nonseekable_open(inode, file);
}

static int biffirq_release(struct inode *inode, struct file *file)
{
   unsigned long irqflags;
   
   //module_put(THIS_MODULE);
   
   spin_lock_irqsave(&biffirq_lock, irqflags);
   
   _taken_cond_flags &= ~((u32)file->private_data);
   _dev_open--;
   
   spin_unlock_irqrestore(&biffirq_lock, irqflags);
   
   return 0;
}

static ssize_t biffirq_read(struct file* file, char* buf, 
                            size_t count, loff_t *ppos)
{
   unsigned long cpr = 0;
   char c;
   
   if (count > 0) {
      // if ppos is not 0, that's not our first read on this
      // fd, so block until the pin value changes.
      if (*ppos > 0) {
         _pin_val_change_cond_read |= (u32)file->private_data;
         
         wait_event_interruptible(biffirq_wait,
            (0 == (_pin_val_change_cond_read & (u32)file->private_data)));
         
         _pin_val_change_cond_read &= ~((u32)file->private_data);
      }
      
      c = '0' + _pin_val;
      cpr = 1 - copy_to_user(buf, &c, 1);
   }
   
   *ppos = cpr;
   
   return cpr;
}

static unsigned int biffirq_poll(struct file* file, poll_table* wait)
{
   unsigned int mask = POLLIN | POLLRDNORM;
   
   _pin_val_change_cond_poll |= (u32)file->private_data;
   
   wait_event_interruptible(biffirq_wait,
      (0 == (_pin_val_change_cond_poll & (u32)file->private_data)));
   
   _pin_val_change_cond_poll &= ~((u32)file->private_data);
   
   return mask;
}

static const struct file_operations biffirq_fops = {
   .owner      = THIS_MODULE,
   .llseek     = no_llseek,
   .open       = biffirq_open,
   .read       = biffirq_read,
   .poll       = biffirq_poll,
   .release    = biffirq_release,
};

static struct miscdevice biffirq_dev = {
   MISC_DYNAMIC_MINOR,
   BIFFIRQ_NAME,
   &biffirq_fops
};

static int __init biffirq_init_module(void)
{  
   int err = 0;
   
   printk(KERN_INFO BIFFIRQ_DRV_NAME ": registering.\n");
   
   err = configure_irq_pin(BIFFIRQ_PIN_NUM);
   
   if (err)
      goto fault;
   
   err = misc_register(&biffirq_dev);
   if (err) {
      free_irq(BIFFIRQ_IRQ_NUM, (void*)(&irq_handler));
   }

fault:   
   return err;
}

static void __exit biffirq_cleanup_module(void)
{
   printk(KERN_INFO BIFFIRQ_DRV_NAME ": deregistering...\n");   
   free_irq(BIFFIRQ_IRQ_NUM, (void*)(&irq_handler));
   misc_deregister(&biffirq_dev);
   
   // turn off reporting of the line status change interrupt
   clear_bit_at_port_and_verify8(MODEM_IRQ_ENABLE, (1<<3));
}

module_init(biffirq_init_module);
module_exit(biffirq_cleanup_module);
MODULE_AUTHOR("Georg Kaindl <gkaindl@mac.com>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Bifferboard value-change interrupt"
                     " on GPIO pin 12.");